from .abstract_callback import AbstractCallback
from delira import get_backends

if 'TORCH' in get_backends():
    from torch.optim.lr_scheduler import ReduceLROnPlateau, CosineAnnealingLR, \
        ExponentialLR, LambdaLR, MultiStepLR, StepLR

    class DefaultPyTorchSchedulerCallback(AbstractCallback):
        """
        Implements a Callback, which `at_epoch_end` function is suitable for most
        schedulers

        """

        def __init__(self, *args, **kwargs):
            """

            Parameters
            ----------
            *args :
                Arbitrary Positional Arguments
            **kwargs :
                Arbitrary Keyword Arguments

            """
            super().__init__()

            self.scheduler = None

        def at_epoch_end(self, trainer, **kwargs):
            """
            Executes a single scheduling step

            Parameters
            ----------
            trainer : :class:`PyTorchNetworkTrainer`
                the trainer class, which can be changed
            **kwargs :
                additional keyword arguments

            Returns
            -------
            :class:`PyTorchNetworkTrainer`
                modified trainer

            """
            self.scheduler.step(epoch=kwargs.get("curr_epoch", None))
            return {}

    class ReduceLROnPlateauCallback(DefaultPyTorchSchedulerCallback):
        """
        Wraps PyTorch's `ReduceLROnPlateau` Scheduler as Callback

        """

        def __init__(self, optimizer, mode='min', factor=0.1, patience=10,
                     verbose=False, threshold=1e-4, threshold_mode='rel',
                     cooldown=0, min_lr=0, eps=1e-8):
            """

            Parameters
            ----------
            optimizer : Optimizer
                Wrapped optimizer.
            mode : str
                One of `min`, `max`. In `min` mode, lr will
                be reduced when the quantity monitored has stopped
                decreasing; in `max` mode it will be reduced when the
                quantity monitored has stopped increasing. Default: 'min'.
            factor : float
                Factor by which the learning rate will be
                reduced. new_lr = lr * factor. Default: 0.1.
            patience : int
                Number of epochs with no improvement after
                which learning rate will be reduced. For example, if
                `patience = 2`, then we will ignore the first 2 epochs
                with no improvement, and will only decrease the LR after the
                3rd epoch if the loss still hasn't improved then.
                Default: 10.
            verbose : bool
                If ``True``, prints a message to stdout for
                each update. Default: ``False``.
            threshold : float
                Threshold for measuring the new optimum,
                to only focus on significant changes. Default: 1e-4.
            threshold_mode : string
                One of `rel`, `abs`. In `rel` mode,
                dynamic_threshold = best * ( 1 + threshold ) in 'max'
                mode or best * ( 1 - threshold ) in `min` mode.
                In `abs` mode, dynamic_threshold = best + threshold in
                `max` mode or best - threshold in `min` mode. Default: 'rel'.
            cooldown : int
                Number of epochs to wait before resuming
                normal operation after lr has been reduced. Default: 0.
            min_lr : float or list
                A scalar or a list of scalars. A
                lower bound on the learning rate of all param groups
                or each group respectively. Default: 0.
            eps : float
                Minimal decay applied to lr. If the difference
                between new and old lr is smaller than eps, the update is
                ignored. Default: 1e-8

            """
            super().__init__()
            self.scheduler = ReduceLROnPlateau(optimizer, mode, factor, patience,
                                               verbose, threshold, threshold_mode,
                                               cooldown, min_lr, eps)

        def at_epoch_end(self, trainer,
                         **kwargs):
            """
            Executes a single scheduling step

            Parameters
            ----------
            trainer : :class:`PyTorchNetworkTrainer`
                the trainer class, which can be changed
            kwargs :
                additional keyword arguments

            Returns
            -------
            :class:`PyTorchNetworkTrainer`
                modified trainer

            """
            metrics = kwargs.get("val_metrics", {}).get(kwargs.get("val_score_key",
                                                                   None))
            self.scheduler.step(metrics=metrics)

            return {}

    class CosineAnnealingLRCallback(DefaultPyTorchSchedulerCallback):
        """
        Wraps PyTorch's `CosineAnnealingLR` Scheduler as callback

        """

        def __init__(self, optimizer, T_max, eta_min=0, last_epoch=-1):
            """

            Parameters
            ----------
            optimizer : optimizer
                Wrapped optimizer.
            T_max : int
                Maximum number of iterations.
            eta_min : float
                Minimum learning rate. Default: 0.
            last_epoch : int
                The index of last epoch. Default: -1.

            """
            super().__init__()

            self.scheduler = CosineAnnealingLR(optimizer, T_max, eta_min,
                                               last_epoch)

    class ExponentialLRCallback(DefaultPyTorchSchedulerCallback):
        """
        Wraps PyTorch's `ExponentialLR` Scheduler as callback

        """

        def __init__(self, optimizer, gamma, last_epoch=-1):
            """

            Parameters
            ----------
            optimizer : Optimizer
                Wrapped optimizer.
            gamma : float
                Multiplicative factor of learning rate decay.
            last_epoch : int
                The index of last epoch. Default: -1.

            """
            super().__init__()

            self.scheduler = ExponentialLR(optimizer, gamma, last_epoch)

    class LambdaLRCallback(DefaultPyTorchSchedulerCallback):
        """
        Wraps PyTorch's `LambdaLR` Scheduler as callback

        """

        def __init__(self, optimizer, lr_lambda, last_epoch=-1):
            """

            Parameters
            ----------
            optimizer : Optimizer
                Wrapped optimizer.
            lr_lambda : function or list
                A function which computes a multiplicative
                factor given an integer parameter epoch, or a list of such
                functions, one for each group in optimizer.param_groups.
            last_epoch : int
                The index of last epoch. Default: -1.

            """
            super().__init__()

            self.scheduler = LambdaLR(optimizer, lr_lambda, last_epoch)

    class MultiStepLRCallback(DefaultPyTorchSchedulerCallback):
        """
        Wraps PyTorch's `MultiStepLR` Scheduler as callback

        """

        def __init__(self, optimizer, milestones, gamma=0.1, last_epoch=-1):
            """

            Parameters
            ----------
            optimizer : Optimizer
                Wrapped optimizer.
            milestones : list
                List of epoch indices. Must be increasing.
            gamma : float
                Multiplicative factor of learning rate decay.
                Default: 0.1.
            last_epoch : int
                The index of last epoch. Default: -1.

            """
            super().__init__()

            self.scheduler = MultiStepLR(
                optimizer, milestones, gamma, last_epoch)

    class StepLRCallback(DefaultPyTorchSchedulerCallback):
        """
        Wraps PyTorch's `StepLR` Scheduler as callback

        """

        def __init__(self, optimizer, step_size, gamma=0.1, last_epoch=-1):
            """

            Parameters
            ----------
            optimizer : Optimizer
                Wrapped optimizer.
            step_size : int
                Period of learning rate decay.
            gamma :float
                Multiplicative factor of learning rate decay.
                Default: 0.1.
            last_epoch : int
                The index of last epoch. Default: -1

            """
            super().__init__()

            self.scheduler = StepLR(optimizer, step_size, gamma, last_epoch)



from trixi.util import Config
from nested_lookup import nested_lookup
from typing import Optional


class LookupConfig(Config):
    """
    Helper class to have nested lookups in all subdicts of Config

    """

    def nested_get(self, key, *args, **kwargs):
        """
        Returns all occurances of ``key`` in ``self`` and subdicts

        Parameters
        ----------
        key : str
            the key to search for
        *args :
            positional arguments to provide default value
        **kwargs :
            keyword arguments to provide default value

        Raises
        ------
        KeyError
            Multiple Values are found for key 
            (unclear which value should be returned)
            OR
            No Value was found for key and no default value was given

        Returns
        -------
        Any
            value corresponding to key (or default if value was not found)

        """

        if "." in key:
            return self[key]
        results = nested_lookup(key, self)
        if len(results) > 1:
            raise KeyError("Multiple Values found for key %s" % key)
        elif len(results) == 0:
            if "default" in kwargs:
                return kwargs["default"]
            elif args:
                return args[0]
            else:
                raise KeyError("No Value found for key %s" % key)
        else:
            return results[0]

    def __setattr__(self, key, value):
        """
        Modified to automatically convert `dict` to LookupConfig.

        Parameters
        ----------
        key : str
            the key which should be set
        value : Any
            the corresponding value to set the key to.

        """

        if type(value) == dict or type(value) == Config:
            new_config = LookupConfig()
            new_config.update(value, deep=False)
            super().__setattr__(key, new_config)
        else:
            super().__setattr__(key, value)



from random import shuffle
from collections import OrderedDict
from .abstract_sampler import AbstractSampler
from ..dataset import AbstractDataset


class SequentialSampler(AbstractSampler):
    """
    Implements Sequential Sampling from whole Dataset
    """
    def __init__(self, indices):
        """

        Parameters
        ----------
        indices : list
            list of classes each sample belongs to. List index corresponds to
            data index and the value at a certain index indicates the
            corresponding class
        """
        super().__init__()

        self._indices = list(range(len(indices)))
        self._global_index = 0

    def _get_indices(self, n_indices):
        """
        Actual Sampling

        Parameters
        ----------
        n_indices : int
            number of indices to return

        Raises
        ------
        StopIteration : If end of dataset reached

        Returns
        -------
        list
            list of sampled indices
        """
        if self._global_index >= len(self._indices):
            self._global_index = 0
            raise StopIteration

        new_global_idx = self._global_index + n_indices

        # If we reach end, make batch smaller
        if new_global_idx >= len(self._indices):
            new_global_idx = len(self._indices)

        idxs = list(range(self._global_index, new_global_idx))
        self._global_index = new_global_idx

        return idxs

    def __len__(self):
        return len(self._indices)


class PrevalenceSequentialSampler(AbstractSampler):
    """
    Implements Per-Class Sequential sampling and ensures same
    number of samples per batch for each class; If out of samples for one
    class: restart at first sample

    """

    def __init__(self, indices, shuffle_batch=True):
        """

        Parameters
        ----------
        indices : list
            list of classes each sample belongs to. List index corresponds to
            data index and the value at a certain index indicates the
             corresponding class
        shuffle_batch : bool
            if False: indices per class will be returned in a sequential way
            (first: indices belonging to class 1, second: indices belonging
            to class 2 etc.)
            if True: indices will be sampled in a sequential way per class and
            sampled indices will be shuffled
        """
        super().__init__()

        _indices = {}
        _global_idxs = {}
        for idx, class_idx in enumerate(indices):
            class_idx = int(class_idx)
            if class_idx in _indices.keys():
                _indices[class_idx].append(idx)
            else:
                _indices[class_idx] = [idx]
                _global_idxs[class_idx] = 0

        # sort classes after descending number of elements
        ordered_dict = OrderedDict()

        length = 0
        for k in sorted(_indices, key=lambda k: len(_indices[k]),
                        reverse=True):
            ordered_dict[k] = _indices[k]
            length += len(_indices[k])

        self._length = length

        self._indices = ordered_dict
        self._n_classes = len(_indices.keys())
        self._global_idxs = _global_idxs
        self._global_total_idx = 0

        self._shuffle = shuffle_batch

    @classmethod
    def from_dataset(cls, dataset: AbstractDataset, **kwargs):
        indices = range(len(dataset))
        labels = [dataset[idx]['label'] for idx in indices]
        return cls(labels, **kwargs)

    def _get_indices(self, n_indices):
        """
        Actual Sampling

        Parameters
        ----------
        n_indices : int
            number of indices to return

        Raises
        ------
        StopIteration : If end of class indices is reached

        Returns
        -------
        list
            list of sampled indices
        """
        samples_per_class = int(n_indices / self._n_classes)

        if self._global_total_idx >= len(self._indices):
            self._global_total_idx = 0
            raise StopIteration

        _samples = []

        for key, idx_list in self._indices.items():
            if self._global_idxs[key] >= len(idx_list):
                self._global_idxs[key] = 0

            new_global_idx = self._global_idxs[key] + samples_per_class

            if new_global_idx >= len(idx_list):
                new_global_idx = len(idx_list)

            _samples += list(range(self._global_idxs[key], new_global_idx))
            self._global_idxs[key] = new_global_idx

        for key, idx_list in self._indices.items():
            if len(_samples) >= n_indices:
                break

            if self._global_idxs[key] >= len(idx_list):
                self._global_idxs[key] = 0

            new_global_idx = self._global_idxs[key] + 1

            _samples += list(range(self._global_idxs[key], new_global_idx))
            self._global_idxs[key] = new_global_idx

        self._global_total_idx += n_indices

        if self._shuffle:
            shuffle(_samples)

        return _samples

    def __len__(self):
        return self._length


class StoppingPrevalenceSequentialSampler(AbstractSampler):
    """
    Implements Per-Class Sequential sampling and ensures same
    number of samples per batch for each class; Stops if all samples of
    first class have been sampled

    """
    def __init__(self, indices, shuffle_batch=True):
        """

        Parameters
        ----------
        indices : list
            list of classes each sample belongs to. List index corresponds to
            data index and the value at a certain index indicates the
             corresponding class
        shuffle_batch : bool
            if False: indices per class will be returned in a sequential way
            (first: indices belonging to class 1, second: indices belonging
            to class 2 etc.)
            if True: indices will be sampled in a sequential way per class and
            sampled indices will be shuffled
        """
        super().__init__()

        _indices = {}
        _global_idxs = {}
        for idx, class_idx in enumerate(indices):
            class_idx = int(class_idx)
            if class_idx in _indices.keys():
                _indices[class_idx].append(idx)
            else:
                _indices[class_idx] = [idx]
                _global_idxs[class_idx] = 0

        # sort classes after descending number of elements
        ordered_dict = OrderedDict()

        length = float('inf')
        for k in sorted(_indices, key=lambda k: len(_indices[k]),
                        reverse=True):
            ordered_dict[k] = _indices[k]
            length = min(length, len(_indices[k]))

        self._length = length

        self._indices = ordered_dict
        self._n_classes = len(_indices.keys())
        self._global_idxs = _global_idxs

        self._shuffle = shuffle_batch

    @classmethod
    def from_dataset(cls, dataset: AbstractDataset):
        indices = range(len(dataset))
        labels = [dataset[idx]['label'] for idx in indices]
        return cls(labels)

    def _get_indices(self, n_indices):
        """
        Actual Sampling

        Parameters
        ----------
        n_indices : int
            number of indices to return

        Raises
        ------
        StopIteration : If end of class indices is reached for one class

        Returns
        -------
        list
            list of sampled indices

        """
        samples_per_class = int(n_indices/self._n_classes)
        _samples = []

        for key, idx_list in self._indices.items():
            if self._global_idxs[key] >= len(idx_list):
                self._global_idxs[key] = 0
                raise StopIteration

            new_global_idx = self._global_idxs[key] + samples_per_class

            if new_global_idx >= len(idx_list):
                new_global_idx = len(idx_list)

            _samples += list(range(self._global_idxs[key], new_global_idx))
            self._global_idxs[key] = new_global_idx

        for key, idx_list in self._indices.items():
            if len(_samples) >= n_indices:
                break

            if self._global_idxs[key] >= len(idx_list):
                self._global_idxs[key] = 0

            new_global_idx = self._global_idxs[key] + 1

            _samples += list(range(self._global_idxs[key], new_global_idx))
            self._global_idxs[key] = new_global_idx

        if self._shuffle:
            shuffle(_samples)

        return _samples

    def __len__(self):
        return self._length



# Adapted from https://github.com/jaxony/unet-pytorch/blob/master/model.py

from delira import get_backends

if "TORCH" in get_backends():
    import torch
    import torch.nn.functional as F
    from torch.nn import init
    import logging
    from ..abstract_network import AbstractPyTorchNetwork


    class UNet2dPyTorch(AbstractPyTorchNetwork):
        """
        The :class:`UNet2dPyTorch` is a convolutional encoder-decoder neural
        network.
        Contextual spatial information (from the decoding,
        expansive pathway) about an input tensor is merged with
        information representing the localization of details
        (from the encoding, compressive pathway).

        Notes
        -----
        Differences to the original paper:

            * padding is used in 3x3 convolutions to prevent loss of border pixels
            * merging outputs does not require cropping due to (1)
            * residual connections can be used by specifying ``merge_mode='add'``
            * if non-parametric upsampling is used in the decoder pathway (
                specified by upmode='upsample'), then an additional 1x1 2d
                convolution occurs after upsampling to reduce channel
                dimensionality by a factor of 2. This channel halving happens
                with the convolution in the tranpose convolution (specified by
                ``upmode='transpose'``)

        References
        ----------
        https://arxiv.org/abs/1505.04597

        See Also
        --------
        :class:`UNet3dPyTorch`

        """

        def __init__(self, num_classes, in_channels=1, depth=5,
                    start_filts=64, up_mode='transpose',
                    merge_mode='concat'):
            """

            Parameters
            ----------
            num_classes : int
                number of output classes
            in_channels : int
                number of channels for the input tensor (default: 1)
            depth : int
                number of MaxPools in the U-Net (default: 5)
            start_filts : int
                number of convolutional filters for the first conv (affects all
                other conv-filter numbers too; default: 64)
            up_mode : str
                type of upconvolution. Must be one of ['transpose', 'upsample']
                if 'transpose':
                    Use transpose convolution for upsampling
                if 'upsample':
                    Use bilinear Interpolation for upsampling (no additional
                    trainable parameters)
                default: 'transpose'
            merge_mode : str
                mode of merging the two paths (with and without pooling). Must
                be one of ['merge', 'add']
                if 'merge':
                    Concatenates along the channel dimension (Original UNet)
                if 'add':
                    Adds both tensors (Residual behaviour)
                default: 'merge'

            """

            super().__init__()

            if up_mode in ('transpose', 'upsample'):
                self.up_mode = up_mode
            else:
                raise ValueError("\"{}\" is not a valid mode for "
                                "upsampling. Only \"transpose\" and "
                                "\"upsample\" are allowed.".format(up_mode))

            if merge_mode in ('concat', 'add'):
                self.merge_mode = merge_mode
            else:
                raise ValueError("\"{}\" is not a valid mode for"
                                "merging up and down paths. "
                                "Only \"concat\" and "
                                "\"add\" are allowed.".format(up_mode))

            # NOTE: up_mode 'upsample' is incompatible with merge_mode 'add'
            if self.up_mode == 'upsample' and self.merge_mode == 'add':
                raise ValueError("up_mode \"upsample\" is incompatible "
                                "with merge_mode \"add\" at the moment "
                                "because it doesn't make sense to use "
                                "nearest neighbour to reduce "
                                "depth channels (by half).")

            self.num_classes = num_classes
            self.in_channels = in_channels
            self.start_filts = start_filts
            self.depth = depth

            self.down_convs = []
            self.up_convs = []

            self.conv_final = None

            self._build_model(num_classes, in_channels, depth, start_filts)

            self.reset_params()

        @staticmethod
        def weight_init(m):
            """
            Initializes weights with xavier_normal and bias with zeros

            Parameters
            ----------
            m : torch.nn.Module
                module to initialize

            """
            if isinstance(m, torch.nn.Conv2d):
                init.xavier_normal_(m.weight)
                init.constant_(m.bias, 0)

        def reset_params(self):
            """
            Initialize all parameters

            """
            for i, m in enumerate(self.modules()):
                self.weight_init(m)

        def forward(self, x):
            """
            Feed tensor through network

            Parameters
            ----------
            x : torch.Tensor

            Returns
            -------
            torch.Tensor
                Prediction

            """
            encoder_outs = []

            # encoder pathway, save outputs for merging
            for i, module in enumerate(self.down_convs):
                x, before_pool = module(x)
                encoder_outs.append(before_pool)

            for i, module in enumerate(self.up_convs):
                before_pool = encoder_outs[-(i + 2)]
                x = module(before_pool, x)

            # No softmax is used. This means you need to use
            # torch.nn.CrossEntropyLoss is your training script,
            # as this module includes a softmax already.
            x = self.conv_final(x)
            return x

        @staticmethod
        def closure(model, data_dict: dict, optimizers: dict, criterions={},
                    metrics={}, fold=0, **kwargs):
            """
            closure method to do a single backpropagation step


            Parameters
            ----------
            model : :class:`ClassificationNetworkBasePyTorch`
                trainable model
            data_dict : dict
                dictionary containing the data
            optimizers : dict
                dictionary of optimizers to optimize model's parameters
            criterions : dict
                dict holding the criterions to calculate errors
                (gradients from different criterions will be accumulated)
            metrics : dict
                dict holding the metrics to calculate
            fold : int
                Current Fold in Crossvalidation (default: 0)
            **kwargs:
                additional keyword arguments

            Returns
            -------
            dict
                Metric values (with same keys as input dict metrics)
            dict
                Loss values (with same keys as input dict criterions)
            list
                Arbitrary number of predictions as torch.Tensor

            Raises
            ------
            AssertionError
                if optimizers or criterions are empty or the optimizers are not
                specified

            """

            assert (optimizers and criterions) or not optimizers, \
                "Criterion dict cannot be emtpy, if optimizers are passed"

            loss_vals = {}
            metric_vals = {}
            total_loss = 0

            # choose suitable context manager:
            if optimizers:
                context_man = torch.enable_grad

            else:
                context_man = torch.no_grad

            with context_man():

                inputs = data_dict.pop("data")
                preds = model(inputs)

                if data_dict:

                    for key, crit_fn in criterions.items():
                        _loss_val = crit_fn(preds, *data_dict.values())
                        loss_vals[key] = _loss_val.detach()
                        total_loss += _loss_val

                    with torch.no_grad():
                        for key, metric_fn in metrics.items():
                            metric_vals[key] = metric_fn(
                                preds, *data_dict.values())

            if optimizers:
                optimizers['default'].zero_grad()
                # perform loss scaling via apex if half precision is enabled
                with optimizers["default"].scale_loss(total_loss) as scaled_loss:
                    scaled_loss.backward()
                optimizers['default'].step()

            else:

                # add prefix "val" in validation mode
                eval_loss_vals, eval_metrics_vals = {}, {}
                for key in loss_vals.keys():
                    eval_loss_vals["val_" + str(key)] = loss_vals[key]

                for key in metric_vals:
                    eval_metrics_vals["val_" + str(key)] = metric_vals[key]

                loss_vals = eval_loss_vals
                metric_vals = eval_metrics_vals

            for key, val in {**metric_vals, **loss_vals}.items():
                logging.info({"value": {"value": val.item(), "name": key,
                                        "env_appendix": "_%02d" % fold
                                        }})

            logging.info({'image_grid': {"images": inputs, "name": "input_images",
                                        "env_appendix": "_%02d" % fold}})

            logging.info({'image_grid': {"images": preds,
                                        "name": "predicted_images",
                                        "env_appendix": "_%02d" % fold}})

            return metric_vals, loss_vals, [preds]

        def _build_model(self, num_classes, in_channels=3, depth=5,
                        start_filts=64):
            """
            Builds the actual model

            Parameters
            ----------
            num_classes : int
                number of output classes
            in_channels : int
                number of channels for the input tensor (default: 1)
            depth : int
                number of MaxPools in the U-Net (default: 5)
            start_filts : int
                number of convolutional filters for the first conv (affects all
                other conv-filter numbers too; default: 64)

            Notes
            -----
            The Helper functions and classes are defined within this function
            because ``delira`` offers a possibility to save the source code
            along the weights to completely recover the network without needing
            a manually created network instance and these helper functions have
            to be saved too.

            """

            def conv3x3(in_channels, out_channels, stride=1,
                        padding=1, bias=True, groups=1):
                return torch.nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=3,
                    stride=stride,
                    padding=padding,
                    bias=bias,
                    groups=groups)

            def upconv2x2(in_channels, out_channels, mode='transpose'):
                if mode == 'transpose':
                    return torch.nn.ConvTranspose2d(
                        in_channels,
                        out_channels,
                        kernel_size=2,
                        stride=2)
                else:
                    # out_channels is always going to be the same
                    # as in_channels
                    return torch.nn.Sequential(
                        torch.nn.Upsample(mode='bilinear', scale_factor=2),
                        conv1x1(in_channels, out_channels))

            def conv1x1(in_channels, out_channels, groups=1):
                return torch.nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=1,
                    groups=groups,
                    stride=1)

            class DownConv(torch.nn.Module):
                """
                A helper Module that performs 2 convolutions and 1 MaxPool.
                A ReLU activation follows each convolution.
                """

                def __init__(self, in_channels, out_channels, pooling=True):
                    super(DownConv, self).__init__()

                    self.in_channels = in_channels
                    self.out_channels = out_channels
                    self.pooling = pooling

                    self.conv1 = conv3x3(self.in_channels, self.out_channels)
                    self.conv2 = conv3x3(self.out_channels, self.out_channels)

                    if self.pooling:
                        self.pool = torch.nn.MaxPool2d(kernel_size=2, stride=2)

                def forward(self, x):
                    x = F.relu(self.conv1(x))
                    x = F.relu(self.conv2(x))
                    before_pool = x
                    if self.pooling:
                        x = self.pool(x)
                    return x, before_pool

            class UpConv(torch.nn.Module):
                """
                A helper Module that performs 2 convolutions and 1 UpConvolution.
                A ReLU activation follows each convolution.
                """

                def __init__(self, in_channels, out_channels,
                            merge_mode='concat', up_mode='transpose'):
                    super(UpConv, self).__init__()

                    self.in_channels = in_channels
                    self.out_channels = out_channels
                    self.merge_mode = merge_mode
                    self.up_mode = up_mode

                    self.upconv = upconv2x2(self.in_channels, self.out_channels,
                                            mode=self.up_mode)

                    if self.merge_mode == 'concat':
                        self.conv1 = conv3x3(
                            2 * self.out_channels, self.out_channels)
                    else:
                        # num of input channels to conv2 is same
                        self.conv1 = conv3x3(self.out_channels,
                                            self.out_channels)
                    self.conv2 = conv3x3(self.out_channels, self.out_channels)

                def forward(self, from_down, from_up):
                    from_up = self.upconv(from_up)
                    if self.merge_mode == 'concat':
                        x = torch.cat((from_up, from_down), 1)
                    else:
                        x = from_up + from_down
                    x = F.relu(self.conv1(x))
                    x = F.relu(self.conv2(x))
                    return x

            outs = in_channels
            # create the encoder pathway and add to a list
            for i in range(depth):
                ins = self.in_channels if i == 0 else outs
                outs = start_filts * (2 ** i)
                pooling = True if i < depth - 1 else False

                down_conv = DownConv(ins, outs, pooling=pooling)
                self.down_convs.append(down_conv)

            # create the decoder pathway and add to a list
            # - careful! decoding only requires depth-1 blocks
            for i in range(depth - 1):
                ins = outs
                outs = ins // 2
                up_conv = UpConv(ins, outs, up_mode=self.up_mode,
                                merge_mode=self.merge_mode)
                self.up_convs.append(up_conv)

            self.conv_final = conv1x1(outs, num_classes)

            # add the list of modules to current module
            self.down_convs = torch.nn.ModuleList(self.down_convs)
            self.up_convs = torch.nn.ModuleList(self.up_convs)

        @staticmethod
        def prepare_batch(batch: dict, input_device, output_device):
            """
            Helper Function to prepare Network Inputs and Labels (convert them to
            correct type and shape and push them to correct devices)

            Parameters
            ----------
            batch : dict
                dictionary containing all the data
            input_device : torch.device
                device for network inputs
            output_device : torch.device
                device for network outputs

            Returns
            -------
            dict
                dictionary containing data in correct type and shape and on correct
                device

            """
            return_dict = {"data": torch.from_numpy(batch.pop("data")).to(
                input_device).to(torch.float)}

            for key, vals in batch.items():
                if key == "label" and len(vals.shape) == 4:
                    vals = vals[:, 0]  # remove first axis if to many axis
                                    # (channel dimension)
                return_dict[key] = torch.from_numpy(vals).to(output_device).to(
                    torch.long)

            return return_dict


    class UNet3dPyTorch(AbstractPyTorchNetwork):
        """
        The :class:`UNet3dPyTorch` is a convolutional encoder-decoder neural
        network.
        Contextual spatial information (from the decoding,
        expansive pathway) about an input tensor is merged with
        information representing the localization of details
        (from the encoding, compressive pathway).

        Notes
        -----
        Differences to the original paper:
            * Working on 3D data instead of 2D slices
            * padding is used in 3x3x3 convolutions to prevent loss of border
                pixels
            * merging outputs does not require cropping due to (1)
            * residual connections can be used by specifying ``merge_mode='add'``
            * if non-parametric upsampling is used in the decoder pathway (
                specified by upmode='upsample'), then an additional 1x1x1 3d
                convolution occurs after upsampling to reduce channel
                dimensionality by a factor of 2. This channel halving happens
                with the convolution in the tranpose convolution (specified by
                ``upmode='transpose'``)

        References
        ----------
        https://arxiv.org/abs/1505.04597

        See Also
        --------
        :class:`UNet2dPyTorch`

        """

        def __init__(self, num_classes, in_channels=3, depth=5,
                    start_filts=64, up_mode='transpose',
                    merge_mode='concat'):
            """

            Parameters
            ----------
            num_classes : int
                number of output classes
            in_channels : int
                number of channels for the input tensor (default: 1)
            depth : int
                number of MaxPools in the U-Net (default: 5)
            start_filts : int
                number of convolutional filters for the first conv (affects all
                other conv-filter numbers too; default: 64)
            up_mode : str
                type of upconvolution. Must be one of ['transpose', 'upsample']
                if 'transpose':
                    Use transpose convolution for upsampling
                if 'upsample':
                    Use trilinear Interpolation for upsampling (no additional
                    trainable parameters)
                default: 'transpose'
            merge_mode : str
                mode of merging the two paths (with and without pooling). Must
                be one of ['merge', 'add']
                if 'merge':
                    Concatenates along the channel dimension (Original UNet)
                if 'add':
                    Adds both tensors (Residual behaviour)
                default: 'merge'

            """
            super().__init__()

            if up_mode in ('transpose', 'upsample'):
                self.up_mode = up_mode
            else:
                raise ValueError("\"{}\" is not a valid mode for "
                                "upsampling. Only \"transpose\" and "
                                "\"upsample\" are allowed.".format(up_mode))

            if merge_mode in ('concat', 'add'):
                self.merge_mode = merge_mode
            else:
                raise ValueError("\"{}\" is not a valid mode for"
                                "merging up and down paths. "
                                "Only \"concat\" and "
                                "\"add\" are allowed.".format(up_mode))

            # NOTE: up_mode 'upsample' is incompatible with merge_mode 'add'
            if self.up_mode == 'upsample' and self.merge_mode == 'add':
                raise ValueError("up_mode \"upsample\" is incompatible "
                                "with merge_mode \"add\" at the moment "
                                "because it doesn't make sense to use "
                                "nearest neighbour to reduce "
                                "depth channels (by half).")

            self.num_classes = num_classes
            self.in_channels = in_channels
            self.start_filts = start_filts
            self.depth = depth

            self.down_convs = []
            self.up_convs = []
            self.conv_final = None

            self._build_model(num_classes, in_channels, depth, start_filts)

            self.reset_params()

        @staticmethod
        def weight_init(m):
            """
            Initializes weights with xavier_normal and bias with zeros

            Parameters
            ----------
            m : torch.nn.Module
                module to initialize

            """

            if isinstance(m, torch.nn.Conv3d):
                init.xavier_normal_(m.weight)
                init.constant_(m.bias, 0)

        def reset_params(self):
            """
            Initialize all parameters

            """
            for i, m in enumerate(self.modules()):
                self.weight_init(m)

        def forward(self, x):
            """
            Feed tensor through network

            Parameters
            ----------
            x : torch.Tensor

            Returns
            -------
            torch.Tensor
                Prediction

            """
            encoder_outs = []

            # encoder pathway, save outputs for merging
            for i, module in enumerate(self.down_convs):
                x, before_pool = module(x)
                encoder_outs.append(before_pool)

            for i, module in enumerate(self.up_convs):
                before_pool = encoder_outs[-(i + 2)]
                x = module(before_pool, x)

            # No softmax is used. This means you need to use
            # torch.nn.CrossEntropyLoss is your training script,
            # as this module includes a softmax already.
            x = self.conv_final(x)
            return x

        @staticmethod
        def closure(model, data_dict: dict, optimizers: dict, criterions={},
                    metrics={}, fold=0, **kwargs):
            """
            closure method to do a single backpropagation step


            Parameters
            ----------
            model : :class:`ClassificationNetworkBasePyTorch`
                trainable model
            data_dict : dict
                dictionary containing the data
            optimizers : dict
                dictionary of optimizers to optimize model's parameters
            criterions : dict
                dict holding the criterions to calculate errors
                (gradients from different criterions will be accumulated)
            metrics : dict
                dict holding the metrics to calculate
            fold : int
                Current Fold in Crossvalidation (default: 0)
            **kwargs:
                additional keyword arguments

            Returns
            -------
            dict
                Metric values (with same keys as input dict metrics)
            dict
                Loss values (with same keys as input dict criterions)
            list
                Arbitrary number of predictions as torch.Tensor

            Raises
            ------
            AssertionError
                if optimizers or criterions are empty or the optimizers are not
                specified

            """

            assert (optimizers and criterions) or not optimizers, \
                "Criterion dict cannot be emtpy, if optimizers are passed"

            loss_vals = {}
            metric_vals = {}
            total_loss = 0

            # choose suitable context manager:
            if optimizers:
                context_man = torch.enable_grad

            else:
                context_man = torch.no_grad

            with context_man():

                inputs = data_dict.pop("data")
                preds = model(inputs)

                if data_dict:

                    for key, crit_fn in criterions.items():
                        _loss_val = crit_fn(preds, *data_dict.values())
                        loss_vals[key] = _loss_val.detach()
                        total_loss += _loss_val

                    with torch.no_grad():
                        for key, metric_fn in metrics.items():
                            metric_vals[key] = metric_fn(
                                preds, *data_dict.values())

            if optimizers:
                optimizers['default'].zero_grad()
                # perform loss scaling via apex if half precision is enabled
                with optimizers["default"].scale_loss(total_loss) as scaled_loss:
                    scaled_loss.backward()
                optimizers['default'].step()

            else:

                # add prefix "val" in validation mode
                eval_loss_vals, eval_metrics_vals = {}, {}
                for key in loss_vals.keys():
                    eval_loss_vals["val_" + str(key)] = loss_vals[key]

                for key in metric_vals:
                    eval_metrics_vals["val_" + str(key)] = metric_vals[key]

                loss_vals = eval_loss_vals
                metric_vals = eval_metrics_vals

            for key, val in {**metric_vals, **loss_vals}.items():
                logging.info({"value": {"value": val.item(), "name": key,
                                        "env_appendix": "_%02d" % fold
                                        }})

            slicing_dim = inputs.size(2) // 2  # visualize slice in mid of volume

            logging.info({'image_grid': {"inputs": inputs[:, :, slicing_dim, ...],
                                        "name":
                                            "input_images",
                                        "env_appendix": "_%02d" % fold}})

            logging.info({'image_grid': {"results": preds[:, :, slicing_dim, ...],
                                        "name":
                                            "predicted_images",
                                        "env_appendix": "_%02d" % fold}})

            return metric_vals, loss_vals, [preds]

        def _build_model(self, num_classes, in_channels=3, depth=5,
                        start_filts=64):
            """
            Builds the actual model

            Parameters
            ----------
            num_classes : int
                number of output classes
            in_channels : int
                number of channels for the input tensor (default: 1)
            depth : int
                number of MaxPools in the U-Net (default: 5)
            start_filts : int
                number of convolutional filters for the first conv (affects all
                other conv-filter numbers too; default: 64)

            Notes
            -----
            The Helper functions and classes are defined within this function
            because ``delira`` offers a possibility to save the source code
            along the weights to completely recover the network without needing
            a manually created network instance and these helper functions have
            to be saved too.

            """

            def conv3x3x3(in_channels, out_channels, stride=1,
                        padding=1, bias=True, groups=1):
                return torch.nn.Conv3d(
                    in_channels,
                    out_channels,
                    kernel_size=3,
                    stride=stride,
                    padding=padding,
                    bias=bias,
                    groups=groups)

            def upconv2x2x2(in_channels, out_channels, mode='transpose'):
                if mode == 'transpose':
                    return torch.nn.ConvTranspose3d(
                        in_channels,
                        out_channels,
                        kernel_size=2,
                        stride=2)
                else:
                    # out_channels is always going to be the same
                    # as in_channels
                    return torch.nn.Sequential(
                        torch.nn.Upsample(mode='trilinear', scale_factor=2),
                        conv1x1x1(in_channels, out_channels))

            def conv1x1x1(in_channels, out_channels, groups=1):
                return torch.nn.Conv3d(
                    in_channels,
                    out_channels,
                    kernel_size=1,
                    groups=groups,
                    stride=1)

            class DownConv(torch.nn.Module):
                """
                A helper Module that performs 2 convolutions and 1 MaxPool.
                A ReLU activation follows each convolution.
                """

                def __init__(self, in_channels, out_channels, pooling=True):
                    super(DownConv, self).__init__()

                    self.in_channels = in_channels
                    self.out_channels = out_channels
                    self.pooling = pooling

                    self.conv1 = conv3x3x3(self.in_channels, self.out_channels)
                    self.conv2 = conv3x3x3(self.out_channels, self.out_channels)

                    if self.pooling:
                        self.pool = torch.nn.MaxPool3d(kernel_size=2, stride=2)

                def forward(self, x):
                    x = F.relu(self.conv1(x))
                    x = F.relu(self.conv2(x))
                    before_pool = x
                    if self.pooling:
                        x = self.pool(x)
                    return x, before_pool

            class UpConv(torch.nn.Module):
                """
                A helper Module that performs 2 convolutions and 1 UpConvolution.
                A ReLU activation follows each convolution.
                """

                def __init__(self, in_channels, out_channels,
                            merge_mode='concat', up_mode='transpose'):
                    super(UpConv, self).__init__()

                    self.in_channels = in_channels
                    self.out_channels = out_channels
                    self.merge_mode = merge_mode
                    self.up_mode = up_mode

                    self.upconv = upconv2x2x2(self.in_channels, self.out_channels,
                                            mode=self.up_mode)

                    if self.merge_mode == 'concat':
                        self.conv1 = conv3x3x3(
                            2 * self.out_channels, self.out_channels)
                    else:
                        # num of input channels to conv2 is same
                        self.conv1 = conv3x3x3(self.out_channels,
                                            self.out_channels)
                    self.conv2 = conv3x3x3(self.out_channels, self.out_channels)

                def forward(self, from_down, from_up):
                    from_up = self.upconv(from_up)
                    if self.merge_mode == 'concat':
                        x = torch.cat((from_up, from_down), 1)
                    else:
                        x = from_up + from_down
                    x = F.relu(self.conv1(x))
                    x = F.relu(self.conv2(x))
                    return x

            outs = in_channels
            # create the encoder pathway and add to a list
            for i in range(depth):
                ins = self.in_channels if i == 0 else outs
                outs = start_filts * (2 ** i)
                pooling = True if i < depth - 1 else False

                down_conv = DownConv(ins, outs, pooling=pooling)
                self.down_convs.append(down_conv)

            # create the decoder pathway and add to a list
            # - careful! decoding only requires depth-1 blocks
            for i in range(depth - 1):
                ins = outs
                outs = ins // 2
                up_conv = UpConv(ins, outs, up_mode=self.up_mode,
                                merge_mode=self.merge_mode)
                self.up_convs.append(up_conv)

            self.conv_final = conv1x1x1(outs, num_classes)

            # add the list of modules to current module
            self.down_convs = torch.nn.ModuleList(self.down_convs)
            self.up_convs = torch.nn.ModuleList(self.up_convs)

        @staticmethod
        def prepare_batch(batch: dict, input_device, output_device):
            """
            Helper Function to prepare Network Inputs and Labels (convert them to
            correct type and shape and push them to correct devices)

            Parameters
            ----------
            batch : dict
                dictionary containing all the data
            input_device : torch.device
                device for network inputs
            output_device : torch.device
                device for network outputs

            Returns
            -------
            dict
                dictionary containing data in correct type and shape and on correct
                device

            """
            return_dict = {"data": torch.from_numpy(batch.pop("data")).to(
                input_device).to(torch.float)}

            for key, vals in batch.items():
                if key == "label" and len(vals.shape) == 5:
                    vals = vals[:, 0]  # remove first axis if to many axis
                    # (channel dimension)
                return_dict[key] = torch.from_numpy(vals).to(output_device).to(
                    torch.long)

            return return_dict



from delira import get_backends

if "TORCH" in get_backends():
    from .classification_network import ClassificationNetworkBasePyTorch
    from .classification_network_3D import VGG3DClassificationNetworkPyTorch

if "TF" in get_backends():
    from .classification_network_tf import ClassificationNetworkBaseTf
    from .ResNet18 import ResNet18



import numpy as np
from skimage.io import imread
from skimage.transform import resize


def default_load_fn_2d(img_file, *label_files, img_shape, n_channels=1):
    """
    loading single 2d sample with arbitrary number of samples

    Parameters
    ----------
    img_file : string
        path to image file
    label_files : list of strings
        paths to label files
    img_shape : iterable
        shape of image
    n_channels : int
        number of image channels

    Returns
    -------
    numpy.ndarray
        image
    Any:
        labels

    """
    img = imread(img_file, n_channels == 1)
    labels = [np.loadtxt(_file).reshape(1).astype(np.float32) for _file in
              label_files]
    img = resize(img, img_shape, mode='reflect', anti_aliasing=True)
    img = np.reshape(img, (*img_shape, n_channels))
    img = img.transpose((len(img_shape), *range(len(img_shape))))
    img = img.astype(np.float32)
    # return img, labels

    result_dict = {"data": img}

    for idx, label in enumerate(labels):
        result_dict["label_%d" % idx] = label

    return result_dict



import logging
import tensorflow as tf
logger = logging.getLogger(__name__)


def save_checkpoint(file: str, model=None):
    """
    Save model's parameters contained in it's graph

    Parameters
    ----------
    file : str
        filepath the model should be saved to
    model : TfNetwork
        the model which should be saved
    """
    tf.train.Saver().save(model._sess, file)


def load_checkpoint(file: str, model=None):
    """
    Loads a saved model

    Parameters
    ----------
    file : str
        filepath to a file containing a saved model
    model : TfNetwork
        the model which should be loaded
    """

    # following operation adds AssignVariableOps to the graph, keep an eye on this for memory leak
    tf.train.Saver().restore(model._sess, file)



import logging
import numpy as np
import typing
import inspect
from batchgenerators.dataloading import SlimDataLoaderBase, \
    MultiThreadedAugmenter
from batchgenerators.transforms import AbstractTransform
from .dataset import AbstractDataset, BaseCacheDataset, BaseLazyDataset, \
    ConcatDataset
from .data_loader import BaseDataLoader
from .load_utils import default_load_fn_2d
from .sampler import SequentialSampler, AbstractSampler
from ..utils.decorators import make_deprecated

logger = logging.getLogger(__name__)


class BaseDataManager(object):
    """
    Class to Handle Data
    Creates Dataset , Dataloader and BatchGenerator

    """

    def __init__(self, data, batch_size, n_process_augmentation,
                 transforms, sampler_cls=SequentialSampler,
                 sampler_kwargs={},
                 data_loader_cls=None, dataset_cls=None,
                 load_fn=default_load_fn_2d, from_disc=True, **kwargs):
        """

        Parameters
        ----------
        data : str or Dataset
            if str: Path to data samples
            if dataset: Dataset
        batch_size : int
            Number of samples per batch
        n_process_augmentation : int
            Number of processes for augmentations
        transforms :
            Data transformations for augmentation
        sampler_cls : AbstractSampler
            class defining the sampling strategy
        sampler_kwargs : dict
            keyword arguments for sampling
        data_loader_cls : subclass of SlimDataLoaderBase
            DataLoader class
        dataset_cls : subclass of AbstractDataset
            Dataset class
        load_fn : function
            function to load simple sample
        from_disc : bool
            whether or not to load data from disc just the time it is needed
        **kwargs :
            other keyword arguments (needed for dataloading and passed to
            dataset_cls)

        Raises
        ------
        AssertionError
            * `data_loader_cls` is not :obj:`None` and not a subclass of
            `SlimDataLoaderBase`
            * `dataset_cls` is not :obj:`None` and not a subclass of
            :class:`.AbstractDataset`

        See Also
        --------
        :class:`AbstractDataset`

        """
        
        # Instantiate Hidden variables for property access
        self._batch_size = None
        self._n_process_augmentation = None
        self._transforms = None
        self._data_loader_cls = None
        self._dataset = None
        self._sampler = None

        # set actual values to properties
        self.batch_size = batch_size

        self.n_process_augmentation = n_process_augmentation
        self.transforms = transforms

        if data_loader_cls is None:
            logger.info("No DataLoader Class specified. Using BaseDataLoader")
            data_loader_cls = BaseDataLoader
        else:
            assert inspect.isclass(data_loader_cls), \
                "data_loader_cls must be class not instance of class"
            assert issubclass(data_loader_cls, SlimDataLoaderBase), \
                "dater_loader_cls must be subclass of SlimDataLoaderBase"

        self.data_loader_cls = data_loader_cls

        if isinstance(data, AbstractDataset):
            self.dataset = data
        else:

            if dataset_cls is None:
                if from_disc:
                    dataset_cls = BaseLazyDataset
                else:
                    dataset_cls = BaseCacheDataset

                logger.info("No DataSet Class specified. Using %s instead" %
                            dataset_cls.__name__)

            else:
                assert issubclass(dataset_cls, AbstractDataset), \
                    "dataset_cls must be subclass of AbstractDataset"

            self.dataset = dataset_cls(data, load_fn, **kwargs)
        
        assert inspect.isclass(sampler_cls) and issubclass(sampler_cls,
                                                            AbstractSampler)
        self.sampler = sampler_cls.from_dataset(self.dataset, **sampler_kwargs)

    def get_batchgen(self, seed=1):
        """
        Create DataLoader and Batchgenerator

        Parameters
        ----------
        seed : int
            seed for Random Number Generator

        Returns
        -------
        MultiThreadedAugmenter
            Batchgenerator

        Raises
        ------
        AssertionError
            :attr:`BaseDataManager.n_batches` is smaller than or equal to zero

        """
        assert self.n_batches > 0

        data_loader = self.data_loader_cls(self.dataset,
                                           batch_size=self.batch_size,
                                           num_batches=self.n_batches,
                                           seed=seed,
                                           sampler=self.sampler
                                           )

        return MultiThreadedAugmenter(data_loader, self.transforms,
                                      self.n_process_augmentation,
                                      num_cached_per_queue=2,
                                      seeds=self.n_process_augmentation*[seed])

    def get_subset(self, indices):
        """
        Returns a Subset of the current datamanager based on given indices
        
        Parameters
        ----------
        indices : iterable
            valid indices to extract subset from current dataset

        Returns
        -------
        :class:`BaseDataManager`
            manager containing the subset
        
        """

        subset_kwargs = {
            "batch_size": self.batch_size,
            "n_process_augmentation": self.n_process_augmentation,
            "transforms": self.transforms,
            "sampler_cls": self.sampler.__class__,
            "data_loader_cls": self.data_loader_cls,
            "dataset_cls": None,
            "load_fn": None,
            "from_disc": True
        }

        return self.__class__(self.dataset.get_subset(indices), **subset_kwargs)

    def update_state_from_dict(self, new_state: dict):
        """
        Updates internal state and therfore the behavior from dict.
        If a key is not specified, the old attribute value will be used
        
        Parameters
        ----------
        new_state : dict
            The dict to update the state from.
            Valid keys are:

                * ``batch_size``
                * ``n_process_augmentation``
                * ``data_loader_cls``
                * ``sampler``
                * ``sampling_kwargs``
                * ``transforms``

            If a key is not specified, the old value of the corresponding 
            attribute will be used

        Raises
        ------
        KeyError
            Invalid keys are specified
        
        """

        # update batch_size if specified
        self.batch_size = new_state.pop("batch_size", self.batch_size)
        # update n_process_augmentation if specified
        self.n_process_augmentation = new_state.pop("n_process_augmentation",
                                                    self.n_process_augmentation)
        # update data_loader_cls if specified
        self.data_loader_cls = new_state.pop("data_loader_cls",
                                                self.data_loader_cls)
        # update
        new_sampler = new_state.pop("sampler", None)
        if new_sampler is not None:
            self.sampler = new_sampler.from_dataset(
                self.dataset,
                **new_state.pop("sampling_kwargs", {}))
        self.transforms = new_state.pop("transforms", self.transforms)

        if new_state:
            raise KeyError("Invalid Keys in new_state given: %s"
                            % (','.join(map(str, new_state.keys()))))

    @make_deprecated("BaseDataManager.get_subset")
    def train_test_split(self, *args, **kwargs):
        """
        Calls :method:`AbstractDataset.train_test_split` and returns 
        a manager for each subset with same configuration as current manager

        .. deprecation:: 0.3
            method will be removed in next major release

        Parameters
        ----------
        *args : 
            positional arguments for 
            ``sklearn.model_selection.train_test_split``
        **kwargs :
            keyword arguments for 
            ``sklearn.model_selection.train_test_split``
        
        """

        trainset, valset = self.dataset.train_test_split(*args, **kwargs)

        subset_kwargs = {
            "batch_size": self.batch_size,
            "n_process_augmentation": self.n_process_augmentation,
            "transforms": self.transforms,
            "sampler_cls": self.sampler.__class__,
            "data_loader_cls": self.data_loader_cls,
            "dataset_cls": None,
            "load_fn": None,
            "from_disc": True
        }

        train_mgr = self.__class__(trainset, **subset_kwargs)
        val_mgr = self.__class__(valset, **subset_kwargs)

        return train_mgr, val_mgr

    @property
    def batch_size(self):
        """
        Property to access the batchsize
        
        Returns
        -------
        int
            the batchsize
        """

        return self._batch_size

    @batch_size.setter
    def batch_size(self, new_batch_size):
        """
        Setter for current batchsize, casts to int before setting the attribute
        
        Parameters
        ----------
        new_batch_size : int, Any
            the new batchsize; should be int but can be of any type that can be 
            casted to an int
        
        """

        self._batch_size = int(new_batch_size)
       
    @property
    def n_process_augmentation(self):
        """
        Property to access the number of augmentation processes
        
        Returns
        -------
        int
            number of augmentation processes
        """

        return self._n_process_augmentation

    @n_process_augmentation.setter
    def n_process_augmentation(self, new_process_number):
        """
        Setter for number of augmentation processes, casts to int before setting
        the attribute
        
        Parameters
        ----------
        new_process_number : int, Any
            new number of augmentation processes; should be int but can be of 
            any type that can be casted to an int
        
        """

        self._n_process_augmentation = int(new_process_number)

    @property
    def transforms(self):
        """
        Property to access the current data transforms
        
        Returns
        -------
        None, ``AbstractTransform``
            The transformation, can either be None or an instance of 
            ``AbstractTransform``
        """

        return self._transforms
    
    @transforms.setter
    def transforms(self, new_transforms):
        """
        Setter for data transforms, assert if transforms are of valid type 
        (either None or instance of ``AbstractTransform``)
        
        Parameters
        ----------
        new_transforms : None, ``AbstractTransform``
            the new transforms
        
        """

        assert new_transforms is None or isinstance(new_transforms,
                                                    AbstractTransform)

        self._transforms = new_transforms

    @property
    def data_loader_cls(self):
        """
        Property to access the current data loader class
        
        Returns
        -------
        type
            Subclass of ``SlimDataLoaderBase``
        """

        return self._data_loader_cls

    @data_loader_cls.setter
    def data_loader_cls(self, new_loader_cls):
        """
        Setter for current data loader class, asserts if class is of valid type
        (must be a class and a subclass of ``SlimDataLoaderBase``)
        
        Parameters
        ----------
        new_loader_cls : type
            the new data loader class
        
        """

        assert inspect.isclass(new_loader_cls) and issubclass(new_loader_cls,
                                                        SlimDataLoaderBase)
        self._data_loader_cls = new_loader_cls

    @property
    def dataset(self):
        """
        Property to access the current dataset
        
        Returns
        -------
        :class:`AbstractDataset`
            the current dataset

        """

        return self._dataset

    @dataset.setter
    def dataset(self, new_dataset):
        """
        Setter for new dataset
        
        Parameters
        ----------
        new_dataset : :class:`AbstractDataset`
            
        """

        assert isinstance(new_dataset, AbstractDataset)
        self._dataset = new_dataset

    @property
    def sampler(self):
        """
        Property to access the current sampler

        Returns
        -------

        :class:`AbstractSampler`
            the current sampler
        """

        return self._sampler

    @sampler.setter
    def sampler(self, new_sampler):
        """
        Setter for current sampler.
        If a valid class instance is passed, the sampler is simply assigned, if 
        a valid class type is passed, the sampler is created from the dataset
        
        Parameters
        ----------
        new_sampler : :class:`AbstractSampler`, type
            instance or class object of new sampler
        
        Raises
        ------
        ValueError
            Neither a valid class instance nor a valid class type is given
        
        """

        if inspect.isclass(new_sampler) and issubclass(new_sampler,
                                                        AbstractSampler):
            self._sampler = new_sampler.from_dataset(self.dataset)

        elif isinstance(new_sampler, AbstractSampler):
            self._sampler = new_sampler

        else:
            raise ValueError("Given Sampler is neither a subclass of \
                            AbstractSampler, nor an instance of a sampler ")
                            
    @property
    def n_samples(self):
        """
        Number of Samples

        Returns
        -------
        int
            Number of Samples

        """
        return len(self.sampler)

    @property
    def n_batches(self):
        """
        Returns Number of Batches based on batchsize,
        number of samples and number of processes

        Returns
        -------
        int
            Number of Batches

        Raises
        ------
        AssertionError
            :attr:`BaseDataManager.n_samples` is smaller than or equal to zero

        """
        assert self.n_samples > 0
        if self.n_process_augmentation == 1:
            n_batches = int(np.floor(self.n_samples / self.batch_size))
        elif self.n_process_augmentation > 1:
            if (self.n_samples / self.batch_size) < self.n_process_augmentation:
                self.n_process_augmentation = 1
                logger.warning('Too few samples for n_process_augmentation={}. '
                               'Forcing n_process_augmentation={} '
                               'instead'.format(self.n_process_augmentation, 1))
            n_batches = int(np.floor(self.n_samples / self.batch_size /
                                     self.n_process_augmentation))
        else:
            raise ValueError('Invalid value for n_process_augmentation')
        return n_batches



import os
import re
from setuptools import find_packages, setup


def resolve_requirements(file):
    requirements = []
    with open(file) as f:
        req = f.read().splitlines()
        for r in req:
            if r.startswith("-r"):
                requirements += resolve_requirements(
                    os.path.join(os.path.dirname(file), r.split(" ")[1]))
            else:
                requirements.append(r)
    return requirements


def read_file(file):
    with open(file) as f:
        content = f.read()
    return content


def find_version(file):
    content = read_file(file)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", content,
                              re.M)
    if version_match:
        return version_match.group(1)


requirements = resolve_requirements(os.path.join(os.path.dirname(__file__),
                                                 'requirements.txt'))

requirements_extra_full = []

requirements_extra_torch = resolve_requirements(os.path.join(
    os.path.dirname(__file__), 'requirements_extra_torch.txt'))
requirements_extra_full += requirements_extra_torch

requirements_extra_tf = resolve_requirements(os.path.join(
    os.path.dirname(__file__), 'requirements_extra_tf.txt'))
requirements_extra_full += requirements_extra_tf

readme = read_file(os.path.join(os.path.dirname(__file__), "README.md"))
license = read_file(os.path.join(os.path.dirname(__file__), "LICENSE"))
delira_version = find_version(os.path.join(os.path.dirname(__file__), "delira",
                                           "__init__.py"))

setup(
    name='delira',
    version=delira_version,
    packages=find_packages(),
    url='https://git.lfb.rwth-aachen.de/Radiology/Delira/',
    test_suite="pytest",
    long_description=readme,
    license=license,
    install_requires=requirements,
    tests_require=["pytest-cov"],
    python_requires=">3.5",
    extras_require={
        "full": requirements_extra_full,
        "torch": requirements_extra_torch,
        "tf": requirements_extra_tf
    }
)



from .abstract_sampler import AbstractSampler


class LambdaSampler(AbstractSampler):
    """
    Implements Arbitrary Sampling methods specified by a function which takes
    the index_list and the number of indices to return

    """
    def __init__(self, indices, sampling_fn):
        """

        Parameters
        ----------
        indices : list
            list of classes each sample belongs to. List index corresponds to
            data index and the value at a certain index indicates the
             corresponding class
        sampling_fn : function
            Actual sampling implementation; must accept an index-list
            and the number of indices to return

        """
        super().__init__()
        self._indices = list(range(len(indices)))

        self._sampling_fn = sampling_fn
        self._global_index = 0

    def _get_indices(self, n_indices):
        """
        Actual Sampling

        Parameters
        ----------
        n_indices : int
            number of indices to return

        Returns
        -------
        list
            list of sampled indices

        Raises
        ------
        StopIteration
            Maximum number of indices sampled

        """

        if self._global_index >= len(self._indices):
            self._global_index = 0
            raise StopIteration

        new_global_idx = self._global_index + n_indices

        # If we reach end, make batch smaller
        if new_global_idx >= len(self._indices):
            new_global_idx = len(self._indices)

        samples = self._sampling_fn(self._indices,
                                    new_global_idx - self._global_index)

        self._global_index = new_global_idx
        return samples

    def __len__(self):
        return len(self._indices)



import logging

file_logger = logging.getLogger(__name__)

from delira import get_backends

if "TORCH" in get_backends():
    import torch.nn as nn
    import torch.nn.functional as F
    from .classification_network import ClassificationNetworkBasePyTorch

    class VGG3DClassificationNetworkPyTorch(ClassificationNetworkBasePyTorch):
        """
        Exemplaric VGG Network for 3D Classification

        Notes
        -----
        The original network has been adjusted to fit for 3D data

        References
        ----------
        https://arxiv.org/abs/1409.1556

        See Also
        --------
        :class:`ClassificationNetworkBasePyTorch`

        """

        def __init__(self, in_channels: int, n_outputs: int, **kwargs):
            """

            Parameters
            ----------
            in_channels : int
                number of input channels
            n_outputs : int
                number of outputs
            **kwargs :
                additional keyword arguments
            """
            super().__init__(in_channels, n_outputs, **kwargs)

        @staticmethod
        def _build_model(in_channels: int, n_outputs: int, **kwargs):
            """
            Helper Function to build the actual model

            Parameters
            ----------
            in_channels : int
                number of input channels
            n_outputs : int
                number of outputs
            **kwargs :
                additional keyword arguments

            Returns
            -------
            torch.nn.Module
                ensembeled model

            """
            class VGGlike3D(nn.Module):
                def __init__(self, in_channels=3, n_outputs=2):
                    super().__init__()
                    self.conv1 = nn.Conv3d(
                        in_channels, 64, 3, stride=2, padding=0)
                    self.conv2 = nn.Conv3d(64, 64, 3, stride=1, padding=0)
                    self.bn1 = nn.BatchNorm3d(64)

                    self.conv3 = nn.Conv3d(64, 128, 3, stride=2, padding=0)
                    self.conv4 = nn.Conv3d(128, 128, 3, stride=1, padding=0)
                    self.bn2 = nn.BatchNorm3d(128)

                    self.conv5 = nn.Conv3d(128, 256, 3, stride=2, padding=0)
                    self.conv6 = nn.Conv3d(256, 256, (1, 3, 3), stride=1,
                                           padding=0)
                    self.bn3 = nn.BatchNorm3d(256)

                    self.pool = nn.AdaptiveMaxPool3d((1, 16, 16))
                    self.fc1 = nn.Linear(in_features=65536, out_features=1024)
                    self.dropout1 = nn.Dropout(p=0.5, inplace=True)
                    self.fc2 = nn.Linear(1024, 64)
                    self.dropout2 = nn.Dropout(p=0.1, inplace=True)
                    self.fc3 = nn.Linear(64, n_outputs)

                def forward(self, x):
                    x = self.conv1(x)
                    x = self.conv2(x)
                    x = self.bn1(x)
                    x = F.leaky_relu(x, inplace=True)

                    x = self.conv3(x)
                    x = self.conv4(x)
                    x = self.bn2(x)
                    x = F.leaky_relu(x, inplace=True)

                    x = self.conv5(x)
                    x = self.conv6(x)
                    x = self.bn3(x)
                    x = F.leaky_relu(x, inplace=True)

                    x = self.pool(x)
                    x = x.view(x.size(0), -1)
                    x = F.leaky_relu(self.dropout1(self.fc1(x)), inplace=True)
                    x = F.leaky_relu(self.dropout2(self.fc2(x)), inplace=True)
                    x = F.softmax(self.fc3(x), dim=1)
                    return x

            _model = VGGlike3D(in_channels=in_channels, n_outputs=n_outputs)
            return _model



from delira import get_backends
import pytest

@pytest.mark.skipif("TORCH" not in get_backends(),
                    reason="No TORCH Backend Installed")
def test_load_save():
    
    from delira.io import torch_load_checkpoint, torch_save_checkpoint
    from delira.models import AbstractPyTorchNetwork
    import torch

    class DummyNetwork(AbstractPyTorchNetwork):
        def __init__(self, in_channels, n_outputs):
            super().__init__(in_channels=in_channels, n_outputs=n_outputs)
            self.net = self._build_model(in_channels, n_outputs)

        def forward(self, x):
            return self.module(x)

        @staticmethod
        def _build_model(in_channels, n_outputs):
            return torch.nn.Sequential(
                torch.nn.Linear(in_channels, 64),
                torch.nn.ReLU(),
                torch.nn.Linear(64, n_outputs)
            )

    net = DummyNetwork(32, 1)
    torch_save_checkpoint("./model.pt", model=net)
    assert torch_load_checkpoint("./model.pt")


if __name__ == '__main__':
    test_load_save()



from ..utils import LookupConfig
import pickle
import yaml

class Parameters(LookupConfig):
    """
    Class Containing all variable and fixed parameters for training and model 
    instantiation

    See Also
    --------
    ``trixi.util.Config``

    """

    def __init__(self, fixed_params={"model": {},
                                     "training": {}},
                 variable_params={"model": {},
                                  "training": {}}):
        """

        Parameters
        ----------
        fixed_params : dict
            fixed parameters (are not variated using hyperparameter search)
        variable_params: dict
            variable parameters (can be variated by a hyperparameter search)
        """

        super().__init__(fixed=fixed_params,
                         variable=variable_params)

    def permute_hierarchy(self):
        """
        switches hierarchy

        Returns
        -------
        Parameters
            the class with a permuted hierarchy

        Raises
        ------
        AttributeError
            if no valid hierarchy is found

        """

        if self.variability_on_top:
            fixed = self.pop("fixed")
            variable = self.pop("variable")

            model = {
                "fixed": fixed.pop("model"),
                "variable": variable.pop("model")}
            training = {"fixed": fixed.pop("training"),
                        "variable": variable.pop("training")}

            self.model = model
            self.training = training

        elif self.training_on_top:
            model = self.pop("model")
            training = self.pop("training")

            fixed = {
                "model": model.pop("fixed"),
                "training": training.pop("fixed")}
            variable = {
                "model": model.pop("variable"),
                "training": training.pop("variable")}

            self.fixed = fixed
            self.variable = variable

        else:
            return AttributeError(
                "%s must either have keys ('model', 'training') or "
                "('fixed', 'variable')" % self.__class__.__name__)

        return self

    def permute_training_on_top(self):
        """
        permutes hierarchy in a way that the training-model hierarchy is on top

        Returns
        -------
        Parameters
            Parameters with permuted hierarchy

        """

        if self.training_on_top:
            return self
        else:
            return self.permute_hierarchy()

    def permute_variability_on_top(self):
        """
        permutes hierarchy in a way that the training-model hierarchy is on top

        Returns
        -------
        Parameters
            Parameters with permuted hierarchy

        """
        if self.variability_on_top:
            return self
        else:
            return self.permute_hierarchy()

    @property
    def hierarchy(self):
        """
        Returns the current hierarchy

        Returns
        -------
        str
            current hierarchy

        """

        if self.variability_on_top:
            hierarchy = "variability\n|\n->\ttraining"
        elif self.training_on_top:
            hierarchy = "training\n|\n->\tvariability"

        else:
            hierarchy = "no valid hierarchy"

        return hierarchy

    def permute_to_hierarchy(self, hierarchy: str):
        """
        Permute hierarchy to match the specified hierarchy

        Parameters
        ----------
        hierarchy : str
            target hierarchy

        Raises
        ------
        ValueError
            Specified hierarchy is invalid

        Returns
        -------
        Parameters
            parameters with proper hierarchy

        """

        if hierarchy == "variability\n|\n->\ttraining":
            return self.permute_training_on_top()
        elif hierarchy == "training\n|\n->\tvariability":
            return self.permute_variability_on_top()

        else:
            raise ValueError("Invalid Hierarchy: %s" % hierarchy)

    @property
    def variability_on_top(self):
        """
        Return whether the variability is on top

        Returns
        -------
        bool
            whether variability is on top

        """

        return hasattr(self, "fixed") and hasattr(self, "variable")

    @property
    def training_on_top(self):
        """
        Return whether the training hierarchy is on top

        Returns
        -------
        bool
            whether training is on top

        """
        return hasattr(self, "model") and hasattr(self, "training")

    def save(self, filepath: str):
        """
        Saves class to given filepath (YAML + Pickle)

        Parameters
        ----------
        filepath : str
            file to save data to

        """

        if not (filepath.endswith(".yaml") or filepath.endswith(".yml")):
            filepath = filepath + ".yml"

        try:
            with open(filepath, "w") as f:
                yaml.dump(self.permute_variability_on_top(), f)

        except TypeError:
            pass

        finally:
            with open(filepath.replace(".yaml", "").replace(".yml", ""),
                      "wb") as f:
                pickle.dump(self, f)
        
    def __str__(self):
        """
        String Representation of class
        
        Returns
        -------
        str
            string representation
        """

        s = "Parameters:\n"
        for k, v in vars(self).items():
            try:
                s += "\t{} = {}\n".format(k, v)
            except TypeError:
                s += "\t{} = {}\n".format(k, v.__class__.__name__)
        return s

    



from delira.data_loading.sampler import LambdaSampler, \
                                        PrevalenceRandomSampler, \
                                        PrevalenceSequentialSampler, \
                                        RandomSampler, \
                                        SequentialSampler, \
                                        StoppingPrevalenceRandomSampler, \
                                        StoppingPrevalenceSequentialSampler, \
                                        WeightedRandomSampler

import numpy as np
from . import DummyDataset

def test_lambda_sampler():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])

    def sampling_fn_a(index_list, n_indices):
        return index_list[:n_indices]

    def sampling_fn_b(index_list, n_indices):
        return index_list[-n_indices:]
    
    sampler_a = LambdaSampler(list(range(len(dset))), sampling_fn_a)
    sampler_b = LambdaSampler(list(range(len(dset))), sampling_fn_b)

    assert sampler_a(15) == list(range(15))
    assert sampler_b(15) == list(range(len(dset) - 15, len(dset)))

def test_prevalence_random_sampler():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])
    
    sampler = PrevalenceRandomSampler.from_dataset(dset)

    for batch_len in [1, 2, 3]:
    
        equal_batch = sampler(batch_len)

        seen_labels = []
        for idx in equal_batch:
            curr_label = dset[idx]["label"]

            if curr_label not in seen_labels:
                seen_labels.append(curr_label)
            else:
                assert False, "Label already seen and labels must be unique. \
                                Batch length: %d" % batch_len


    assert len(sampler(5)) == 5

def test_prevalence_sequential_sampler():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])
    
    sampler = PrevalenceSequentialSampler.from_dataset(dset)

    # ToDo add test considering actual sampling strategy

    assert len(sampler(5)) == 5

def test_random_sampler():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])

    sampler = RandomSampler.from_dataset(dset)
    
    assert len(sampler(250)) == 250

    # checks if labels are all the same (should not happen if random sampled)
    assert len(set([dset[_idx]["label"] for _idx in sampler(301)])) > 1

def test_sequential_sampler():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])

    sampler = SequentialSampler.from_dataset(dset)

    # if sequentially sampled, the first 300 items should have label 0 -> 1 
    # unique element
    assert len(set([dset[_idx]["label"] for _idx in sampler(100)])) == 1
    assert len(sampler(100)) == 100
    # next 100 elements also same label -> next 201 elements: two different 
    # labels
    assert len(set([dset[_idx]["label"] for _idx in sampler(101)])) == 2

def test_stopping_prevalence_random_sampler():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])

    sampler = StoppingPrevalenceRandomSampler.from_dataset(dset)

    try:
        for i in range(121):
            sample = sampler(3)
            assert len(set([dset[_idx]["label"] for _idx in sample])) == 3

        assert False, "Sampler should have raised StopIteration by now"

    except StopIteration:
        assert True

def test_stopping_prevalence_sequential_sampler():
    np.random.seed(1)
    dset = DummyDataset(600, [0.5, 0.3, 0.2])

    sampler = StoppingPrevalenceRandomSampler.from_dataset(dset)

    try:
        for i in range(121):
            sample = sampler(3)
            assert len(set([dset[_idx]["label"] for _idx in sample])) == 3

        assert False, "Sampler should have raised StopIteration by now"

    except StopIteration:
        assert True


if __name__ == '__main__':
    test_lambda_sampler()
    test_prevalence_random_sampler()
    test_prevalence_sequential_sampler()
    test_random_sampler()
    test_sequential_sampler()
    test_stopping_prevalence_random_sampler()
    test_stopping_prevalence_sequential_sampler()



from delira.data_loading import AbstractDataset
import numpy as np

class DummyDataset(AbstractDataset):
    def __init__(self, length=600, class_weights=[0.5, 0.3, 0.2]):
        super().__init__(None, None, None, None)

        assert sum(class_weights) == 1.

        self._data = [np.random.rand(1, 28, 28) for i in range(length)]
        _labels = []
        for idx, weight in enumerate(class_weights):
            _labels += [idx] * int(length*weight)

        self._labels = _labels

    def __getitem__(self, index):
        return {"data": self._data[index], "label": self._labels[index]}

    def __len__(self):
        return len(self._data)


class AbstractCallback(object):
    """
    Implements abstract callback interface.
    All callbacks should be derived from this class

    See Also
    --------
    :class:`AbstractNetworkTrainer`

    """

    def __init__(self, *args, **kwargs):
        """

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        """
        pass

    def at_epoch_begin(self, trainer, **kwargs):
        """
        Function which will be executed at begin of each epoch

        Parameters
        ----------
        trainer : :class:`AbstractNetworkTrainer`
        **kwargs :
            additional keyword arguments

        Returns
        -------
        dict
            modified trainer attributes, where the name must correspond to the 
            trainer's attribute name

        """
        return {}

    def at_epoch_end(self, trainer, **kwargs):
        """
        Function which will be executed at end of each epoch

        Parameters
        ----------
        trainer : :class:`AbstractNetworkTrainer`
        **kwargs :
            additional keyword arguments

        Returns
        -------
        dict
            modified trainer attributes, where the name must correspond to the 
            trainer's attribute name

        """
        return {}



from .parameters import Parameters
from .experiment import AbstractExperiment
from .abstract_trainer import AbstractNetworkTrainer

from delira import get_backends

if "TORCH" in get_backends():
    from .experiment import PyTorchExperiment
    from .pytorch_trainer import PyTorchNetworkTrainer
    from .metrics import AccuracyMetricPyTorch, AurocMetricPyTorch

if "TF" in get_backends():
    from .experiment import TfExperiment
    from .tf_trainer import TfNetworkTrainer





from ..utils import now
from ..data_loading import BaseDataManager, BaseLazyDataset
from delira import __version__ as delira_version
from .parameters import Parameters
from ..models import AbstractNetwork
from .abstract_trainer import AbstractNetworkTrainer
from trixi.experiment import Experiment as TrixiExperiment
import os
import logging
import typing
import numpy as np

import pickle
from abc import abstractmethod
from sklearn.model_selection import KFold, StratifiedKFold, StratifiedShuffleSplit
from sklearn.model_selection import train_test_split
from datetime import datetime
from inspect import signature
from functools import partial

from delira import get_backends

logger = logging.getLogger(__name__)


class AbstractExperiment(TrixiExperiment):
    """
    Abstract Class Representing a single Experiment (must be subclassed for
    each Backend)

    See Also
    --------
    :class:`PyTorchExperiment`

    """
    @abstractmethod
    def __init__(self, n_epochs, *args, **kwargs):
        """

        Parameters
        ----------
        n_epochs : int
            number of epochs to train
        *args :
            positional arguments
        **kwargs :
            keyword arguments
        """
        super().__init__(n_epochs)
        self._run = 0

    @abstractmethod
    def setup(self, *args, **kwargs):
        """
        Abstract Method to setup a :class:`AbstractNetworkTrainer`

        Raises
        ------
        NotImplementedError
            if not overwritten by subclass

        """
        raise NotImplementedError()

    @abstractmethod
    def run(self, train_data: BaseDataManager,
            val_data: typing.Optional[BaseDataManager] = None,
            params: typing.Optional[Parameters] = None,
            **kwargs):
        """
        trains single model

        Parameters
        ----------
        train_data : :class:`BaseDataManager`
            data manager containing the training data
        val_data : :class:`BaseDataManager`
            data manager containing the validation data
        parameters : :class:`Parameters`, optional
            Class containing all parameters (defaults to None).
            If not specified, the parameters fall back to the ones given during
            class initialization

        Raises
        ------
        NotImplementedError
            If not overwritten in subclass

        """

        raise NotImplementedError()

    @abstractmethod
    def test(self,
             params: Parameters,
             network: AbstractNetwork,
             datamgr_test: BaseDataManager,
             trainer_cls=AbstractNetworkTrainer,
             **kwargs):
        """
        Executes prediction for all items in datamgr_test with network

        Parameters
        ----------
        params : :class:`Parameters`
            the parameters to construct a model
        network : :class:'AbstractNetwork'
            the network to train
        datamgr_test : :class:'BaseDataManager'
            holds the test data
        trainer_cls :
            class defining the actual trainer,
            defaults to :class:`AbstractNetworkTrainer`,
            which should be suitable for most cases,
            but can easily be overwritten and exchanged if necessary
        **kwargs :
            holds additional keyword arguments
            (which are completly passed to the trainers init)

        Returns
        -------
        np.ndarray
            predictions from batches
        list of np.ndarray
            labels from batches
        dict
            dictionary containing the mean validation metrics and
            the mean loss values
        """
        raise NotImplementedError()

    def kfold(self, num_epochs: int,
              data: BaseDataManager,
              num_splits=None, shuffle=False, random_seed=None, train_kwargs={},
              test_kwargs={}, **kwargs):
        """
        Runs K-Fold Crossvalidation
        The supported scenario is:

            * passing a single datamanager: the data within the single manager
            will be split and multiple datamanagers will be created holding
            the subsets.

        Parameters
        ----------
        num_epochs : int
            number of epochs to train the model
        data : single :class:`BaseDataManager`
            single datamanager (will be split for crossvalidation)
        num_splits : None or int
            number of splits for kfold
            if None: 10 splits will be validated per default
        shuffle : bool
            whether or not to shuffle indices for kfold
        random_seed : None or int
            random seed used to seed the kfold (if shuffle is true),
            pytorch and numpy
        train_kwargs : dict
            keyword arguments to specify training behavior
        test_kwargs : dict
            keyword arguments to specify testing behavior
        **kwargs :
            additional keyword arguments (completely passed to self.run())

        See Also
        --------
        :method:`BaseDataManager.update_state_from_dict` for valid keys in
            ``train_kwargs`` and ``test_kwargs``

        """

        # set number of splits if not specified
        if num_splits is None:
            num_splits = 10
            logger.warning("num_splits not defined, using default value of \
                            10 splits instead ")

        # extract actual data to be split
        split_data = list(range(len(data.dataset)))

        # instantiate the actual kfold
        fold = KFold(n_splits=num_splits, shuffle=shuffle,
                     random_state=random_seed)

        if random_seed is not None:
            np.random.seed(random_seed)

        # run folds
        for idx, (train_idxs, test_idxs) in enumerate(fold.split(split_data)):

            # extract data from single manager
            train_data = data.get_subset(train_idxs)
            test_data = data.get_subset(test_idxs)

            # update manager behavior for train and test case
            train_data.update_state_from_dict(train_kwargs)
            test_data.update_state_from_dict(test_kwargs)

            self.run(train_data, test_data,
                     num_epochs=num_epochs,
                     fold=idx,
                     **kwargs)

    def stratified_kfold(self, num_epochs: int,
                         data: BaseDataManager,
                         num_splits=None, shuffle=False, random_seed=None,
                         label_key="label", train_kwargs={}, test_kwargs={},
                         **kwargs):
        """
        Runs stratified K-Fold Crossvalidation
        The supported supported scenario is:

            * passing a single datamanager: the data within the single manager
              will be split and multiple datamanagers will be created holding
              the subsets.

        Parameters
        ----------
        num_epochs : int
            number of epochs to train the model
        data : :class:`BaseDataManager`
            single datamanager
            (will be split for crossvalidation)
        num_splits : None or int
            number of splits for kfold
            if None: 10 splits will be validated
        shuffle : bool
            whether or not to shuffle indices for kfold
        random_seed : None or int
            random seed used to seed the kfold (if shuffle is true),
            pytorch and numpy
        label_key : str (default: "label")
            the key to extract the label for stratification from each data
            sample
        train_kwargs : dict
            keyword arguments to specify training behavior
        test_kwargs : dict
            keyword arguments to specify testing behavior
        **kwargs :
            additional keyword arguments (completely passed to self.run())

        See Also
        --------
        :method:`BaseDataManager.update_state_from_dict` for valid keys in
            ``train_kwargs`` and ``test_kwargs``

        """

        if num_splits is None:
            num_splits = 10
            logger.warning("num_splits not defined, using default value of \
                                10 splits instead ")

        if isinstance(data.dataset, BaseLazyDataset):
            logger.warning("A lazy dataset is given for stratified kfold. \
                            Iterating over the dataset to extract labels for \
                            stratification may be a massive overhead")

        split_idxs = list(range(len(data.dataset)))
        split_labels = [data.dataset[_idx][label_key] for _idx in split_idxs]

        fold = StratifiedKFold(n_splits=num_splits, shuffle=shuffle,
                               random_state=random_seed)

        for idx, (train_idxs, test_idxs) in enumerate(fold.split(split_idxs,
                                                                 split_labels)):
            # extract data from single manager
            train_data = data.get_subset(train_idxs)
            test_data = data.get_subset(test_idxs)

            # update manager behavior for train and test case
            train_data.update_state_from_dict(train_kwargs)
            test_data.update_state_from_dict(test_kwargs)
            self.run(train_data, test_data,
                     num_epochs=num_epochs,
                     fold=idx,
                     **kwargs)

    def stratified_kfold_predict(self, num_epochs: int,
                                 data: BaseDataManager,
                                 split_val=0.2,
                                 num_splits=None, shuffle=False, random_seed=None,
                                 label_key="label", train_kwargs={}, test_kwargs={},
                                 **kwargs):
        """
        Runs stratified K-Fold Crossvalidation
        The supported supported scenario is:

            * passing a single datamanager: the data within the single manager
              will be split and multiple datamanagers will be created holding
              the subsets.

        Parameters
        ----------
        num_epochs : int
            number of epochs to train the model
        data : :class:`BaseDataManager`
            single datamanager
            (will be split for crossvalidation)
        num_splits : None or int
            number of splits for kfold
            if None: 10 splits will be validated
        shuffle : bool
            whether or not to shuffle indices for kfold
        random_seed : None or int
            random seed used to seed the kfold (if shuffle is true),
            pytorch and numpy
        label_key : str (default: "label")
            the key to extract the label for stratification from each data
            sample
        train_kwargs : dict
            keyword arguments to specify training behavior
        test_kwargs : dict
            keyword arguments to specify testing behavior
        **kwargs :
            additional keyword arguments (completely passed to self.run())

        See Also
        --------
        :method:`BaseDataManager.update_state_from_dict` for valid keys in
            ``train_kwargs`` and ``test_kwargs``

        """

        metrics_val = {}
        outputs = {}
        labels = {}

        if num_splits is None:
            num_splits = 10
            logger.warning("num_splits not defined, using default value of \
                                10 splits instead ")

        if isinstance(data.dataset, BaseLazyDataset):
            logger.warning("A lazy dataset is given for stratified kfold. \
                            Iterating over the dataset to extract labels for \
                            stratification may be a massive overhead")

        split_idxs = list(range(len(data.dataset)))
        split_labels = [data.dataset[_idx][label_key] for _idx in split_idxs]

        fold = StratifiedKFold(n_splits=num_splits, shuffle=shuffle,
                               random_state=random_seed)

        for idx, (_train_idxs, test_idxs) in enumerate(fold.split(split_idxs,
                                                                  split_labels)):
            # extract data from single manager
            _train_data = data.get_subset(_train_idxs)
            _split_idxs = list(range(len(_train_data.dataset)))
            _split_labels = [_train_data.dataset[_idx][label_key]
                             for _idx in _split_idxs]

            val_fold = StratifiedShuffleSplit(n_splits=1,
                                              test_size=split_val,
                                              random_state=random_seed)

            for train_idxs, val_idx in val_fold.split(_split_idxs, _split_labels):
                train_data = _train_data.get_subset(train_idxs)
                val_data = _train_data.get_subset(val_idx)

            test_data = data.get_subset(test_idxs)

            # update manager behavior for train and test case
            train_data.update_state_from_dict(train_kwargs)
            val_data.update_state_from_dict(test_kwargs)
            test_data.update_state_from_dict(test_kwargs)
            model = self.run(train_data, val_data,
                             num_epochs=num_epochs,
                             fold=idx,
                             **kwargs)

            _outputs, _labels, _metrics_val = self.test(
                self.params, model, test_data)

            outputs[str(idx)] = _outputs
            labels[str(idx)] = _labels
            metrics_val[str(idx)] = _metrics_val

        return outputs, labels, metrics_val

    def __str__(self):
        """
        Converts :class:`AbstractExperiment` to string representation

        Returns
        -------
        str
            representation of class

        """
        s = "Experiment:\n"
        for k, v in vars(self).items():
            s += "\t{} = {}\n".format(k, v)
        return s

    def __call__(self, *args, **kwargs):
        """
        Call :meth:`AbstractExperiment.run`

        Parameters
        ----------
        *args :
            positional arguments
        **kwargs :
            keyword arguments

        Returns
        -------
        :class:`AbstractNetworkTrainer`
            trainer of trained network

        """
        return self.run(*args, **kwargs)

    @abstractmethod
    def save(self):
        """
        Saves the Whole experiments

        Raises
        ------
        NotImplementedError
            If not overwritten in subclass

        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def load(file_name):
        """
        Loads whole experiment

        Parameters
        ----------
        file_name : str
            file_name to load the experiment from

        Raises
        ------
        NotImplementedError
            if not overwritten in subclass

        """
        raise NotImplementedError()

if "TORCH" in get_backends():

    import torch
    from .train_utils import create_optims_default_pytorch
    from .pytorch_trainer import PyTorchNetworkTrainer as PTNetworkTrainer
    from ..models import AbstractPyTorchNetwork

    class PyTorchExperiment(AbstractExperiment):
        """
        Single Experiment for PyTorch Backend

        See Also
        --------
        :class:`AbstractExperiment`

        """

        def __init__(self,
                     params: Parameters,
                     model_cls: AbstractPyTorchNetwork,
                     name=None,
                     save_path=None,
                     val_score_key=None,
                     optim_builder=create_optims_default_pytorch,
                     checkpoint_freq=1,
                     trainer_cls=PTNetworkTrainer,
                     **kwargs
                     ):
            """

            Parameters
            ----------
            params : :class:`Parameters`
                the training and model parameters
            model_cls :
                the class to instantiate models
            name : str
                the experiment's name, 
                default: None -> "UnnamedExperiment"
            save_path : str
                the path to save the experiment to
                (a date-time signature will be appended),
                default: None -> Use current working dir
            val_score_key : str or None
                key to access the metric to monitor for model
                selection and callbacks (often starts with "val_")
            optim_builder : function
                function returning a dictionary of optimizers
                defaults to :function:`create_optims_default_pytorch`
            checkpoint_freq : int
                save checkpoint after each n epochs
                (if set to 1, checkpoints will be saved after each epoch,
                if set to 2, checkpoints will be saved after each
                2 epochs etc.)
            trainer_cls :
                class defining the actual trainer,
                defaults to :class:`PyTorchNetworkTrainer`,
                which should be suitable for most cases,
                but can easily be overwritten and exchanged if necessary
            **kwargs :
                additional keyword arguments

            """

            if isinstance(params, str):
                with open(params, "rb") as f:
                    params = pickle.load(f)

            n_epochs = params.nested_get("num_epochs")
            AbstractExperiment.__init__(self, n_epochs)

            if name is None:
                name = "UnnamedExperiment"
            self.name = name

            if save_path is None:
                save_path = os.path.abspath(".")

            self.save_path = os.path.join(save_path, name,
                                          str(datetime.now().strftime(
                                              "%y-%m-%d_%H-%M-%S")))

            if os.path.isdir(self.save_path):
                logger.warning("Save Path %s already exists")

            os.makedirs(self.save_path, exist_ok=True)

            self.trainer_cls = trainer_cls

            if val_score_key is None and params.nested_get("metrics"):
                val_score_key = sorted(params.nested_get("metrics").keys())[0]

            self.val_score_key = val_score_key

            self.params = params
            self.model_cls = model_cls
            self.kwargs = kwargs
            self._optim_builder = optim_builder
            self.checkpoint_freq = checkpoint_freq
            self._run = 0

            # log HyperParameters
            logger.info({"text": {"text":
                                  str(params) + "\n\tmodel_class = %s"
                                  % model_cls.__class__.__name__}})

        def setup(self, params: Parameters, **kwargs):
            """
            Perform setup of Network Trainer

            Parameters
            ----------
            params : :class:`Parameters`
                the parameters to construct a model and network trainer
            **kwargs :
                keyword arguments

            """
            model_params = params.permute_training_on_top().model

            model_kwargs = {**model_params.fixed, **model_params.variable}

            model = self.model_cls(**model_kwargs)

            training_params = params.permute_training_on_top().training
            criterions = training_params.nested_get("criterions")
            optimizer_cls = training_params.nested_get("optimizer_cls")
            optimizer_params = training_params.nested_get("optimizer_params")
            metrics = training_params.nested_get("metrics")
            lr_scheduler_cls = training_params.nested_get("lr_sched_cls")
            lr_scheduler_params = training_params.nested_get("lr_sched_params")
            return self.trainer_cls(
                network=model,
                save_path=os.path.join(
                    self.save_path,
                    "checkpoints",
                    "run_%02d" % self._run),
                criterions=criterions,
                optimizer_cls=optimizer_cls,
                optimizer_params=optimizer_params,
                metrics=metrics,
                lr_scheduler_cls=lr_scheduler_cls,
                lr_scheduler_params=lr_scheduler_params,
                optim_fn=self._optim_builder,
                save_freq=self.checkpoint_freq,
                **self.kwargs,
                **kwargs
            )

        def run(self,
                train_data: BaseDataManager,
                val_data: typing.Union[BaseDataManager, None],
                params: typing.Optional[Parameters] = None,
                **kwargs):
            """
            trains single model

            Parameters
            ----------
            train_data : BaseDataManager
                holds the trainset
            val_data : BaseDataManager or None
                holds the validation set (if None: Model will not be validated)
            params : :class:`Parameters`
                the parameters to construct a model and network trainer
            **kwargs :
                holds additional keyword arguments
                (which are completly passed to the trainers init)

            Returns
            -------
            :class:`AbstractNetworkTrainer`
                trainer of trained network

            Raises
            ------
            ValueError
                Class has no Attribute ``params`` and no parameters were given as
                function argument

            """

            if params is None:
                if hasattr(self, "params"):
                    params = self.params
                else:
                    raise ValueError("No parameters given")

            else:
                self.params = params

            training_params = params.permute_training_on_top().training

            trainer = self.setup(params, **kwargs)
            self._run += 1
            num_epochs = kwargs.get("num_epochs", training_params.nested_get(
                "num_epochs"))
            return trainer.train(num_epochs, train_data, val_data,
                                 self.val_score_key,
                                 self.kwargs.get("val_score_mode",
                                                 "lowest")
                                 )

        def test(self,
                 params: Parameters,
                 network: AbstractPyTorchNetwork,
                 datamgr_test: BaseDataManager,
                 **kwargs):
            """
            Executes prediction for all items in datamgr_test with network

            Parameters
            ----------
            params : :class:`Parameters`
                the parameters to construct a model
            network : :class:'AbstractPyTorchNetwork'
                the network to train
            datamgr_test : :class:'BaseDataManager'
                holds the test data
            **kwargs :
                holds additional keyword arguments
                (which are completly passed to the trainers init)

            Returns
            -------
            np.ndarray
                predictions from batches
            list of np.ndarray
                labels from batches
            dict
                dictionary containing the mean validation metrics and
                the mean loss values
            """
            # setup trainer with dummy optimization which won't be used!
            trainer = self.trainer_cls(
                network=network,
                save_path=os.path.join(self.save_path, 'test'),
                optimizer_cls=torch.optim.SGD,
                optim_fn=create_optims_default_pytorch,
                criterions=params.nested_get('criterions'),
                optimizer_params=params.nested_get('optimizer_params'),
                **kwargs)

            # testing with batchsize 1 and 1 augmentation processs to
            # avoid dropping of last elements
            orig_num_aug_processes = datamgr_test.n_process_augmentation
            orig_batch_size = datamgr_test.batch_size

            datamgr_test.batch_size = 1
            datamgr_test.n_process_augmentation = 1

            outputs, labels, metrics_val = trainer.predict(
                datamgr_test.get_batchgen(), batch_size=orig_batch_size)

            # reset old values
            datamgr_test.batch_size = orig_batch_size
            datamgr_test.n_process_augmentation = orig_num_aug_processes
            return outputs, labels, metrics_val

        def save(self):
            """
            Saves the Whole experiments

            """
            with open(os.path.join(self.save_path, "experiment.delira.pkl"),
                      "wb") as f:
                pickle.dump(self, f)

            self.params.save(os.path.join(self.save_path, "parameters"))

        @staticmethod
        def load(file_name):
            """
            Loads whole experiment

            Parameters
            ----------
            file_name : str
                file_name to load the experiment from

            """
            with open(file_name, "rb") as f:
                return pickle.load(f)

        def __getstate__(self):
            return vars(self)

        def __setstate__(self, state):
            vars(self).update(state)

        def kfold(self, num_epochs: int,
                  data: typing.Union[typing.List[BaseDataManager],
                                     BaseDataManager],
                  num_splits=None, shuffle=False, random_seed=None,
                  train_kwargs={}, test_kwargs={}, **kwargs):

            if random_seed is not None:
                torch.manual_seed(random_seed)

            super().kfold(num_epochs, data, num_splits, shuffle, random_seed,
                          train_kwargs, test_kwargs, **kwargs)

        def stratified_kfold(self, num_epochs: int,
                             data: BaseDataManager,
                             num_splits=None, shuffle=False, random_seed=None,
                             label_key="label", train_kwargs={}, test_kwargs={},
                             **kwargs):

            if random_seed is not None:
                torch.manual_seed(random_seed)

            super().stratified_kfold(num_epochs, data, num_splits, shuffle,
                                     random_seed, label_key, train_kwargs,
                                     test_kwargs, **kwargs)

if "TF" in get_backends():

    from .tf_trainer import TfNetworkTrainer
    from .train_utils import create_optims_default_tf
    from ..models import AbstractTfNetwork
    from .parameters import Parameters
    import tensorflow as tf

    class TfExperiment(AbstractExperiment):
        """
        Single Experiment for Tf Backend

        See Also
        --------
        :class:`AbstractExperiment`

        """

        def __init__(self,
                     params: typing.Union[Parameters, str],
                     model_cls: AbstractTfNetwork,
                     name=None,
                     save_path=None,
                     val_score_key=None,
                     optim_builder=create_optims_default_tf,
                     checkpoint_freq=1,
                     trainer_cls=TfNetworkTrainer,
                     **kwargs
                     ):

            if isinstance(params, str):
                with open(params, "rb") as f:
                    params = pickle.load(f)

            n_epochs = params.nested_get("num_epochs")
            AbstractExperiment.__init__(self, n_epochs)

            if name is None:
                name = "UnnamedExperiment"
            self.name = name

            if save_path is None:
                save_path = os.path.abspath(".")

            self.save_path = os.path.join(save_path, name,
                                          str(datetime.now().strftime(
                                              "%y-%m-%d_%H-%M-%S")))

            if os.path.isdir(self.save_path):
                logger.warning("Save Path %s already exists")

            os.makedirs(self.save_path, exist_ok=True)

            self.trainer_cls = trainer_cls

            if val_score_key is None and params.nested_get("metrics"):
                val_score_key = sorted(params.nested_get("metrics").keys())[0]

            self.val_score_key = val_score_key

            self.params = params
            self.model_cls = model_cls
            self.kwargs = kwargs
            self._optim_builder = optim_builder
            self.checkpoint_freq = checkpoint_freq
            self._run = 0

            # log HyperParameters
            logger.info({"text": {"text":
                                  str(params) + "\n\tmodel_class = %s"
                                  % model_cls.__class__.__name__}})

        def setup(self, params: Parameters, **kwargs):
            """
            Perform setup of Network Trainer

            Parameters
            ----------
            params : :class:`Parameters`
                the parameters to construct a model and network trainer
            **kwargs :
                keyword arguments

            """

            model_params = params.permute_training_on_top().model

            model_kwargs = {**model_params.fixed, **model_params.variable}

            tf.reset_default_graph()

            model = self.model_cls(**model_kwargs)

            training_params = params.permute_training_on_top().training
            criterions = training_params.nested_get("criterions")
            optimizer_cls = training_params.nested_get("optimizer_cls")
            optimizer_params = training_params.nested_get("optimizer_params")
            metrics = training_params.nested_get("metrics")
            lr_scheduler_cls = training_params.nested_get("lr_sched_cls")
            lr_scheduler_params = training_params.nested_get("lr_sched_params")

            return self.trainer_cls(
                network=model,
                save_path=os.path.join(
                    self.save_path,
                    "checkpoints",
                    "run_%02d" % self._run),
                losses=criterions,
                optimizer_cls=optimizer_cls,
                optimizer_params=optimizer_params,
                metrics=metrics,
                lr_scheduler_cls=lr_scheduler_cls,
                lr_scheduler_params=lr_scheduler_params,
                optim_fn=self._optim_builder,
                save_freq=self.checkpoint_freq,
                **self.kwargs,
                **kwargs
            )

        def run(self,
                train_data: BaseDataManager,
                val_data: typing.Union[BaseDataManager, None],
                params: typing.Optional[Parameters] = None,
                **kwargs):
            """
            trains single model

            Parameters
            ----------
            train_data : BaseDataManager
                holds the trainset
            val_data : BaseDataManager or None
                holds the validation set (if None: Model will not be validated)
            params : :class:`Parameters`
                the parameters to construct a model and network trainer
            **kwargs :
                holds additional keyword arguments
                (which are completly passed to the trainers init)

            Returns
            -------
            :class:`AbstractNetworkTrainer`
                trainer of trained network

            Raises
            ------
            ValueError
                Class has no Attribute ``params`` and no parameters were given as
                function argument
            """

            if params is None:
                if hasattr(self, "params"):
                    params = self.params
                else:
                    raise ValueError("No parameters given")

            else:
                self.params = params

            training_params = params.permute_training_on_top().training

            trainer = self.setup(params, **kwargs)
            self._run += 1
            num_epochs = kwargs.get("num_epochs", training_params.nested_get(
                "num_epochs"))
            return trainer.train(num_epochs, train_data, val_data,
                                 self.val_score_key,
                                 self.kwargs.get("val_score_mode",
                                                 "lowest")
                                 )

        def save(self):
            """
            Saves the Whole experiments

            """
            with open(os.path.join(self.save_path, "experiment.delira.pkl"),
                      "wb") as f:
                pickle.dump(self, f)

                self.params.save(os.path.join(self.save_path, "parameters"))

        @staticmethod
        def load(file_name):
            """
            Loads whole experiment

            Parameters
            ----------
            file_name : str
                file_name to load the experiment from

            """
            with open(file_name, "rb") as f:
                return pickle.load(f)

        def __getstate__(self):
            return vars(self)

        def __setstate__(self, state):
            vars(self).update(state)

        def kfold(self, num_epochs: int,
                  data: typing.Union[typing.List[BaseDataManager],
                                     BaseDataManager],
                  num_splits=None, shuffle=False, random_seed=None,
                  train_kwargs={}, test_kwargs={}, **kwargs):

            if random_seed is not None:
                tf.set_random_seed(random_seed)

            super().kfold(num_epochs, data, num_splits, shuffle, random_seed,
                          train_kwargs, test_kwargs, **kwargs)

        def stratified_kfold(self, num_epochs: int,
                             data: BaseDataManager,
                             num_splits=None, shuffle=False, random_seed=None,
                             label_key="label", train_kwargs={}, test_kwargs={},
                             **kwargs):

            if random_seed is not None:
                tf.set_random_seed(random_seed)

            super().stratified_kfold(num_epochs, data, num_splits, shuffle,
                                     random_seed, label_key, train_kwargs,
                                     test_kwargs, **kwargs)

        def stratified_kfold_predict(self, num_epochs: int,
                                     data: BaseDataManager,
                                     split_val=0.2,
                                     num_splits=None,
                                     shuffle=False,
                                     random_seed=None,
                                     label_key="label",
                                     train_kwargs={}, test_kwargs={},
                                     **kwargs):

            if random_seed is not None:
                tf.set_random_seed(random_seed)

            return super().stratified_kfold_predict(num_epochs, data, split_val, num_splits, shuffle,
                                                    random_seed, label_key, train_kwargs,
                                                    test_kwargs, **kwargs)

        def test(self,
                 params: Parameters,
                 network: AbstractNetwork,
                 datamgr_test: BaseDataManager,
                 **kwargs):
            """
            Executes prediction for all items in datamgr_test with network

            Parameters
            ----------
            params : :class:`Parameters`
                the parameters to construct a model
            network : :class:'AbstractPyTorchNetwork'
                the network to train
            datamgr_test : :class:'BaseDataManager'
                holds the test data
            **kwargs :
                holds additional keyword arguments
                (which are completly passed to the trainers init)

            Returns
            -------
            np.ndarray
                predictions from batches
            list of np.ndarray
                labels from batches
            dict
                dictionary containing the mean validation metrics and
                the mean loss values
            """

            # setup trainer with dummy optimization which won't be used!

            training_params = params.permute_training_on_top().training
            metrics = training_params.nested_get("metrics")

            trainer = self.trainer_cls(
                network=network,
                save_path=os.path.join(self.save_path, 'test'),
                optimizer_cls=training_params.nested_get("optimizer_cls"),
                optim_fn=create_optims_default_tf,
                optimizer_params={},
                metrics=metrics,
                losses={},
                **kwargs)

            # testing with batchsize 1 and 1 augmentation processs to
            # avoid dropping of last elements
            orig_num_aug_processes = datamgr_test.n_process_augmentation
            orig_batch_size = datamgr_test.batch_size

            datamgr_test.batch_size = 1
            datamgr_test.n_process_augmentation = 1

            outputs, labels, metrics_val = trainer.predict(
                datamgr_test.get_batchgen(), batch_size=orig_batch_size)

            # reset old values
            datamgr_test.batch_size = orig_batch_size
            datamgr_test.n_process_augmentation = orig_num_aug_processes
            return outputs, labels, metrics_val
